# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""Routines for converting between coordinate systems, between 2D cartesian and polar,
as well as 3D cartesian, cylindrical, and spherical.

"""
import numpy as _np


def cart2pol(x, y):
    """Converts from cartesian to polar coordinates

    Args:
        x (ndarray): x coordinates
        y (ndarray): y coordinates

    Returns:
        tuple: rho, phi
    """
    rho = _np.sqrt(x ** 2 + y ** 2)
    phi = _np.arctan2(y, x)
    return (rho, phi)


def pol2cart(rho, phi):
    """Converts from polar to cartesian coordinates

    Args:
        rho (ndarray): radial coordinates
        phi (ndarray): azimuthal coordinates

    Returns:
        tuple: x,y
    """
    x = rho * _np.cos(phi)
    y = rho * _np.sin(phi)
    return (x, y)


def vector_pol2cart(Brho, Bphi, phi):
    """Converts Vectors from polar to cartesian coordinates

    Args:
        Brho (ndarray): radial vector component
        Bphi (ndarray): azimuthal vector component
        phi (ndarray): azimuthal coordinates

    Returns:
        tuple: Bx, By
    """
    Bx = Brho * _np.cos(phi) - Bphi * _np.sin(phi)
    By = Brho * _np.sin(phi) + Bphi * _np.cos(phi)
    return Bx, By


def cart2sph(x, y, z):
    """Converts from cartesian to spherical coordinates

    Args:
        x (ndarray): x coordinates
        y (ndarray): y coordinates
        z (ndarray): z coordinates

    Returns:
        tuple: r, theta, phi
    """
    r = _np.sqrt(x ** 2 + y ** 2 + z ** 2)
    phi = _np.arctan2(y, x)

    # Hide the warning for situtations where there is a divide by zero.
    # This returns a NaN in the array, which is ignored for plotting.
    with _np.errstate(divide="ignore", invalid="ignore"):
        theta = _np.arccos(z / r)
    return (r, theta, phi)


def sph2cart(r, theta, phi):
    """Converts from spherical to cartesian coordinates

    Args:
        r (ndarray): radial coordinates
        theta (ndarray): azimuthal angles
        phi (ndarray): polar angle

    Returns:
        tuple: x,y,z
    """
    x = r * _np.sin(theta) * _np.cos(phi)
    y = r * _np.sin(theta) * _np.sin(phi)
    z = r * _np.cos(theta)
    return x, y, z


def vector_sph2cart(Br, Btheta, Bphi, theta, phi):
    """Converts Vectors from spherical to cartesian coordinates

    Args:
        Br (ndarray): radial vector component
        Btheta (ndarray): polar vector component
        Bphi (ndarray): azimuthal vector component
        theta (ndarray): azimuthal angles
        phi (ndarray): polar angle

    Returns:
        tuple: Bx,By,Bz
    """
    Bx = (
        Br * _np.sin(theta) * _np.cos(phi)
        + Btheta * _np.cos(theta) * _np.cos(phi)
        - Bphi * _np.sin(phi)
    )

    By = (
        Br * _np.sin(theta) * _np.sin(phi)
        + Btheta * _np.cos(theta) * _np.sin(phi)
        + Bphi * _np.cos(phi)
    )

    Bz = Br * _np.cos(theta) - Btheta * _np.sin(theta)
    return Bx, By, Bz


def sphere_sph2cart(Br, Btheta, theta, phi):
    """Converts magnetic field of a sphere from spherical to cartesian coordinates

    Args:
        Br (ndarray): radial vector component
        Btheta (ndarray): polar vector component
        theta (ndarray): azimuthal angles
        phi (ndarray): polar angle

    Returns:
        tuple: Bx,By,Bz
    """
    Bx = Br * _np.sin(theta) * _np.cos(phi) + Btheta * _np.cos(theta) * _np.cos(phi)

    By = Br * _np.sin(theta) * _np.sin(phi) + Btheta * _np.cos(theta) * _np.sin(phi)

    Bz = Br * _np.cos(theta) - Btheta * _np.sin(theta)
    return Bx, By, Bz


def get_unit_value_meter(unit):
    """Returns a queried metre unit as a number
    Example:
        factor = get_unit_value_meter('cm')
        print(f"factor for 'cm' is {factor}")

    Args:
        unit (string): SI length unit

    Returns:
        float: SI prefix factor
    """
    si_prefixes = {
        "Ym": 1e24,
        "Zm": 1e21,
        "Em": 1e18,
        "Pm": 1e15,
        "Tm": 1e12,
        "Gm": 1e9,
        "Mm": 1e6,
        "km": 1e3,
        "hm": 1e2,
        "dam": 1e1,
        "m": 1,
        "dm": 1e-1,
        "cm": 1e-2,
        "mm": 1e-3,
        "µm": 1e-6,
        "um": 1e-6,
        "nm": 1e-9,
        "Ang": 1e-10,
        "pm": 1e-12,
        "fm": 1e-15,
        "am": 1e-18,
        "zm": 1e-21,
        "ym": 1e-24,
    }

    return si_prefixes.get(unit, None)


def get_unit_value_tesla(unit):
    """Returns a queried magnetic flux density unit as a number
    Example:
        factor = get_unit_value_meter('mT')
        print(f"factor for 'mT' is {factor}")

    Args:
        unit (string): SI length unit

    Returns:
        float: SI prefix factor
    """

    si_prefixes = {
        "YT": 1e24,
        "ZT": 1e21,
        "ET": 1e18,
        "PT": 1e15,
        "TT": 1e12,
        "GT": 1e9,
        "MT": 1e6,
        "kT": 1e3,
        "hT": 1e2,
        "daT": 1e1,
        "T": 1,
        "dT": 1e-1,
        "cT": 1e-2,
        "mT": 1e-3,
        "µT": 1e-6,
        "uT": 1e-6,
        "nT": 1e-9,
        "pT": 1e-12,
        "fT": 1e-15,
        "aT": 1e-18,
        "zT": 1e-21,
        "yT": 1e-24,
    }

    return si_prefixes.get(unit, None)


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""Routines for Two Dimensional Magnet Classes
"""
import numpy as _np
from ._vector_structs import Point_Array2, Field2
from .global_const import MU0

__all__ = ["grid2D", "get_field_2D", "rotate_points_2D"]


def grid2D(xmax, ymax, **kwargs):
    """Generates grid of x and y points

    Args:
        xmax (float): maximum x value
        ymax (float): maximum y value

    Kwargs:
        num_points (int): Number of points in each direction. Defaults to 100
        xmin (float): minimum x value. Defaults to -xmax
        ymin (float): minimum y value. Defaults to -ymax
        unit (str): unit length. Defaults to 'mm'

    Returns:
        Point_Array2: array of x and y values of shape (num_points, num_points) and associated unit
    """
    num_points = kwargs.pop("num_points", 100)
    xmin = kwargs.pop("xmin", -1 * xmax)
    ymin = kwargs.pop("ymin", -1 * ymax)
    unit = kwargs.pop("unit", "mm")
    NPJ = num_points * 1j
    x, y = _np.mgrid[xmin:xmax:NPJ, ymin:ymax:NPJ]
    return Point_Array2(x, y, unit=unit)


def get_field_2D(Point_Array2):
    """Calculates magnetic field at an array of points due to every instantated
    `Magnet2D` magnet.

    Args:
        Point_Array2 (Point_Array2): array of x,y points and associated unit, defaults to 'mm'

    Returns:
        Field2: array of Bx,By,|B| values and associated unit (defaults to 'T')
    """
    from ..magnets import Magnet2D

    # Empty data structure
    B = _allocate_field_array2(Point_Array2.x, Point_Array2.y)

    for magnet in Magnet2D.instances:
        Bx, By = magnet.get_field(Point_Array2.x, Point_Array2.y)
        B.x += Bx
        B.y += By

    B.calc_norm()
    return B


def rotate_points_2D(x, y, alpha):
    """Counter-clockwise rotation of points x,y

    Rotates 2D coordinates using a rotation matrix

    Args:
        x (ndarray): array of x coordinates
        y (ndarray): array of x coordinates
        alpha (float): rotation angle w.r.t. x-axis

    Returns:
        tuple: (x', y') rotated array of points
    """
    x = _np.atleast_1d(x)
    y = _np.atleast_1d(y)
    if len(x) != len(y):
        raise Exception("Must have same number of points in x and y")

    rot_matrix = _np.array(
        [[_np.cos(alpha), -_np.sin(alpha)], [_np.sin(alpha), _np.cos(alpha)]]
    )
    stacked_points = _np.column_stack((_np.ravel(x), _np.ravel(y)))
    rotated_points = _np.dot(rot_matrix, stacked_points.T)
    x_rotated = rotated_points[0, :]
    y_rotated = rotated_points[1, :]

    return _np.reshape(x_rotated, x.shape), _np.reshape(y_rotated, y.shape)


def gradB_2D(B, x, y):
    """Calculates the magnetic field gradient for a 2D field.

    Args:
        B (Field2): Magnetic field vector
        x (ndarray): x coordinates
        y (ndarray): y coordinates

    Returns:
        Field2: Magnetic field gradient vector
    """

    dB = Field2(_np.zeros_like(B), _np.zeros_like(B))
    Nx = x.shape[0]
    Ny = x.shape[1]
    dx = (x.max() - x.min()) / Nx
    dy = (y.max() - y.min()) / Ny
    dB.x, dB.y = _np.gradient(B, dx, dy)
    dB.calc_norm()
    return dB


def FgradB_2D(B, x, y, chi_m, c):
    """Calculates the magnetic field gradient force for a 2D field.

    Args:
        B (Field2): Magnetic field vector
        x (ndarray): x coordinates
        y (ndarray): y coordinates

    Returns:
        Field2: Magnetic field gradient force vector
    """

    BgB = Field2(_np.zeros_like(B.n), _np.zeros_like(B.n))
    FB = Field2(_np.zeros_like(B.n), _np.zeros_like(B.n))

    dB = gradB_2D(B, x, y)
    BgB.n = dB.n * B.n
    BgB.x = dB.x * B.n
    BgB.y = dB.y * B.n
    FB.n = (1 / MU0) * chi_m * c * BgB.n
    FB.x = (1 / MU0) * chi_m * c * BgB.x
    FB.y = (1 / MU0) * chi_m * c * BgB.y
    return FB


def _allocate_field_array2(x, y):
    """Allocates empty Field2 data structure

    Args:
        x (ndarray): x co-ordinates
        y (ndarray): y co-ordinates

    Returns:
        Field2: Empty data structure
    """

    # Ensure x,y,z are numpy arrays (even of element 1)
    if _np.isscalar(x):
        x = _np.atleast_1d(x)
    if _np.isscalar(y):
        y = _np.atleast_1d(y)

    # Determine array shape:
    if _np.ndim(x) == 2:  # planar slice
        B = Field2(_np.zeros_like(x), _np.zeros_like(x))
    else:  # line or single point
        B = Field2(_np.zeros(max(x.size, y.size)), _np.zeros(max(x.size, y.size)))
    return B


def _get_field_array_shape2(x, y):
    """Allocates empty Field2 data structure

    Args:
        x (ndarray): x co-ordinates
        y (ndarray): y co-ordinates

    Returns:
        Field2: Empty data structure
    """

    # Ensure x,y,z are numpy arrays (even of element 1)
    x = _np.atleast_1d(x)
    y = _np.atleast_1d(y)

    # Determine array shape:
    if _np.ndim(x) == 2:  # planar slice
        array_shape = x.shape
    else:  # line or single point
        array_shape = max(x.size, y.size)
    return array_shape


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
""" INCOMPLETE Routines for creation of known magnet assemblies consisting of Halbachs, quadrupoles, and others. This does not yet work.

"""
import numpy as _np
from ..utils.global_const import PI, PI_2
from matplotlib.path import Path as _Path


def mask_poly(num_sides, apothem, xc, yc, x, y):
    """A

    Args:
        num_sides (int): Number of sides in polygon
        apothem (float): Polygon apothem
        xc (float): Centre of polygon, x
        yc (float): Centre of polygon, y
        x (ndarray): array of x coordinates
        y (ndarray): array of y coordinates

    Returns:
        ndarray: boolean array of masked points
    """

    # Stack x and y arrays
    points = _np.vstack((x.flatten(), y.flatten())).T

    # Array of number of sides
    k = _np.arange(0, num_sides, 1)

    radius = apothem / _np.around(_np.cos(PI / num_sides), 4)

    # Generate polygon vertices
    xv = xc + radius * _np.sin(2 * PI * k / num_sides + _angle_offset(num_sides))
    yv = yc + radius * _np.cos(2 * PI * k / num_sides + _angle_offset(num_sides))

    # Needs to be a list for _Path function below
    poly_verts = _np.vstack((xv, yv)).T.tolist()

    # Create a _Path object following the vertices
    path = _Path(poly_verts)

    # generate a boolean array of points inside the polygon and reshape
    grid = path.contains_points(points)
    grid = grid.reshape(x.shape)
    return grid


def _angle_offset(num_sides):
    """Case structure for setting angle offset

    Args:
        num_sides (int): number of sides in polygon

    Returns:
        float: offset angle in radians
    """
    # Even number of sides
    if num_sides % 2 == 0:
        return {
            6: 0,
        }.get(num_sides, PI / num_sides)
    # odd number of sides
    else:
        return {
            3: PI_2,
            7: PI_2,
        }.get(num_sides, PI_2 / num_sides)


def mask_data_2D(magnet_half_width, num_sides, radius, B, x, y, **kwargs):
    """Applies mask to field vector.

    All field values outside the mask are set to NaN

    Args:
        magnet_half_width (float): Half the magnet width for an array
        num_sides (int): Number of sides
        radius (float): distance from origin to centre of each magnet
        B (Vector2): Magnetic field vector
        x (ndarray): array of x coordinates
        y (ndarray): array of y coordinates
        **kwargs: keyword arguments

    Kwargs:
        center (tuple): centre of mask, defaults to (.0, .0)
        mask_function (function): mask function, defaults to `mask_poly`, can also be
        `radial_mask`

    Returns:
        Vector2: new vector of masked points
    """
    from ..utils._point_structs import Vector2

    apothem = radius - magnet_half_width
    center = kwargs.pop("center", (0.0, 0.0))
    mask_function = kwargs.pop("mask_function", mask_poly)
    xp = center[0]
    yp = center[1]

    grid = mask_function(num_sides, apothem, xp, yp, x, y)
    Bm = Vector2(B.x, B.y)
    Bm.x[~grid] = _np.nan
    Bm.y[~grid] = _np.nan
    Bm.calc_norm()
    return Bm


def radial_mask(radius, xp, yp, x, y):
    """Applies circular mask

    Args:
        radius (float): mask radius
        xp (float): center of circle, x
        yp (float): center of circle, y
        x (ndarray): array of x points
        y (ndarray): array of y points

    Returns:
        array: boolean array
    """
    r_data = _np.sqrt(_np.power(x - xp, 2) + _np.power(y - yp, 2))
    grid = (r_data) <= radius
    return grid


def radial_profile(data, center):
    """Calculates radial profile

    Args:
        data (ndarray): Data to be profiled must be a 2D array
        center (tuple): [description]

    Returns:
        [type]: [description]
    """
    # Generate a grid of index points depending on the shape of the data
    y, x = _np.indices((data.shape))

    # Generate a radial function in units of index
    r = _np.sqrt(_np.power(x - center[0], 2) + _np.power(y - center[1], 2))

    # Cast to integer
    r = r.astype(_np.int)

    # Bin the data
    tbin = _np.bincount(r.ravel(), data.ravel())
    nr = _np.bincount(r.ravel())
    radialprofile = tbin / nr
    return radialprofile


def radial_extraction_2D(magnet_half_width, radius, x, y, B):
    """Returns the radial average of the normal component of 2D vector field

    Args:
        magnet_half_width (float): half the magnet width
        radius (float): Radius
        x (ndarray): array of x points
        y (ndarray): array of y points
        B (Vector2): Field vector

    Returns:
        tuple: array of radial points, array of radial average
    """

    # Get center indices
    index_x = x.shape[0] // 2
    index_y = y.shape[1] // 2

    # Generate radial length in units of index
    radial_length = _np.sqrt(
        _np.power(x[0, -1] - x[index_x, index_y], 2)
        + _np.power(y[0, -1] - y[index_x, index_y], 2)
    )

    # Get radial average of the normalised field
    radial_average = radial_profile(B.n, [index_x, index_y])

    # Calculate the x value for plotting `xpl`
    xpl = _np.linspace(0, radial_length, len(radial_average))

    # Discard data outside the mask
    radial_average = radial_average[xpl <= radius - magnet_half_width]
    xpl = xpl[xpl <= radius - magnet_half_width]
    return xpl, radial_average


def init_magnets(num_magnets=4, b_scale=1, assem_type="halbach"):
    from .. import reset_magnets
    from ..magnets._magnet2D import Rectangle
    from ._routines2D import grid2D

    """Initialise 

    Args:
        N (int, optional): [Number of magnets]. Defaults to 4.
        b_scale (int, optional): [ratio of b/a]. Defaults to 1.
        assem_type (str, optional): [halbach, 2mag, pseudo_quad]. Defaults to 'halbach'.
    Returns:
        mag_prop (dictionary): Magnet properties
        grid_prop (dictionary): Calculation grid (x,y)
    """
    reset_magnets()
    NP = 201
    width = 2
    height = width * b_scale
    Jr = 1.0

    if assem_type == "halbach".lower():
        theta = 90.0
        alpha = 0.0

        # assembly radius
        radius = (width * 2) / (
            2 * _np.tan(PI / num_magnets)
        ) + width * 1.0  # Assembly radius
        Jr = _np.array([Jr * (-1) ** num for num in range(num_magnets)])
        theta = _np.tile(theta, num_magnets)
        alpha = _np.tile(alpha, num_magnets)

        # Halbach - 4 element
        steps = _np.arange(0, num_magnets, 1)

        alpha = steps * 360 / num_magnets
        xc = -radius * _np.cos(_np.deg2rad(steps * 360 / num_magnets))
        yc = radius * _np.sin(_np.deg2rad(steps * 360 / num_magnets))

        for i in range(num_magnets):
            _ = Rectangle(
                width=width,
                height=height,
                Jr=Jr[i],
                center=(xc[i], yc[i]),
                theta=theta[i],
                alpha=alpha[i],
            )
        off = -width / 2
        UP = radius + off
        UPx = UP
        UPy = UP

    else:
        hGap = width / 2
        xc = width / 2 + hGap
        yc = 0
        theta = 0.0
        alpha = 0.0
        N = 2

        #  2 element
        _ = Rectangle(
            width=width, height=height, Jr=Jr, center=(xc, yc), theta=theta, alpha=alpha
        )
        _ = Rectangle(
            width=width,
            height=height,
            Jr=-Jr,
            center=(-xc, yc),
            theta=theta,
            alpha=alpha,
        )
        UPx = hGap
        UPy = width / 2

    x, y = grid2D(-UPx, UPx, -UPy, UPy, NP)

    mag_prop = {
        "width": width,
        "height": height,
        "xc": xc,
        "yc": yc,
        "NP": NP,
        "Jr": Jr,
        "theta": theta,
        "alpha": alpha,
        "N": num_magnets,
        "radius": radius,
        "assem_type": assem_type.lower(),
    }

    grid_prop = {"x": x, "y": y}

    return mag_prop, grid_prop


def calc_magnetic_field(mag_prop, grid_prop):
    from ..utils._routines2D import get_field_2D

    """Run magnetic calculation and mask data to central bore

    Args:
        mag_prop (dictionary): Magnet properties
        grid_prop (dictionary): Calculation grid (x,y)
    
    Updates grid_prop with magnetic field object
    and masked magnetic field object
    """
    width = mag_prop["width"]
    num_magnets = mag_prop["num_magnets"]

    x = grid_prop["x"]
    y = grid_prop["y"]
    B = get_field_2D(x, y)
    if mag_prop["assem_type"].lower() == "halbach":
        mask_radius = mag_prop["radius"]
    else:
        mask_radius = mag_prop["hGap"]
        N = 4

    Bm = mask_data_2D(width, num_magnets, mask_radius, B, x, y)
    grid_prop.update({"B": B, "Bm": Bm})


# def fit_power(N):
#     """Switch case dictionary for returning powers needed
#     to fit radially averaged magnetic field inside a bore.
#     Defaults to 3 (i.e. for fitting third order polynomial)

#     Args:
#         N (int): [Number of magnets]

#     Returns:
#         [int]: [polynomial order for fit]
#     """
#     return{
#            2: 1,
#            4: 1,
#            6: 2
#            }.get(N, 3)

# def fit_radial_profile(mag_prop, grid_prop):
#     """Calculate radial average inside cavity and fit it with an N-order
#     polynomial

#     Args:
#         mag_prop (dict): magnetic properties
#         grid_prop (dict): Grid properties for calculations (x,y, B, Bm)

#     Returns:
#         radial_param (dict): Radial B, BgradB
#     """

#     x = grid_prop['x']
#     y = grid_prop['y']
#     B = grid_prop['B']
#     a = mag_prop['a']
#     N = mag_prop['N']
#     centre = 0


#     xpl,B_ravg = mag.radial_extraction(x,y,B,a,centre)
#     xpl = xpl[np.isfinite(B_ravg)]
#     B_ravg = B_ravg[np.isfinite(B_ravg)]
#     fit_res = np.polyfit(xpl,B_ravg, fit_power(N))
#     B_fit = np.polyval(fit_res, xpl)
#     c = fit_res[-1]
#     m = fit_res[-2]

#     # num_list = [str(num) for num in fit_res]


#     BgB_fit = B_ravg*m

#     radial_param = {"xpl": xpl,
#                   "B_ravg": B_ravg,
#                   "fit_res": fit_res,
#                   "B_fit": B_fit,
#                   'BgB_fit':BgB_fit}


#     if fit_res.shape[0] > 2:
#         m_approx = np.gradient(B_ravg,xpl).mean()
#         B_fit2 = np.polyval( [m_approx, fit_res[-1]], xpl)
#         if fit_res.shape[0] > 3:
#             BgB_fit = B_fit *( 3*xpl**2 * fit_res[0] + 2*xpl * fit_res[1] + fit_res[2])
#         else:
#             BgB_fit = B_fit *( 2*xpl* fit_res[0] + fit_res[1])
#         m_approx = np.gradient(BgB_fit,xpl).mean()
#         BgB_fit2 = B_fit2*m_approx
#         radial_param = {"xpl": xpl,
#                     "B_ravg": B_ravg,
#                     "fit_res": fit_res,
#                     "B_fit": B_fit,
#                     "B_fit2": B_fit2,
#                     'BgB_fit': BgB_fit,
#                     'BgB_fit2': BgB_fit2
#                     }
#     return radial_param


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""3D Plotting routines

This module contains all functions needed to plot 3D contours for 3D magnetic sources.
Unlike the plot2D module, here plotly is used as the backend.

TODO:
    * Update __str__ and __repr__ for polyhedra
"""
from ..magnets._magnet_base import Registry
from ..magnets import *
from ..utils import grid3D, slice3D, get_field_3D
from ..utils.global_const import PI, MAG_TOL
import numpy as _np
import plotly.graph_objects as _go


__all__ = [
    "plot_magnet",
    "slice_plot",
    "slice_quickplot",
    "volume_plot",
    "volume_quickplot",
]


class Polyhedron(Registry):
    """Encodes magnet dimensions for drawing a polyhedon on 3D plots

    Polyhedra:
        Cuboid
        Cylinder
        Sphere
        Mesh
    """

    # Tolerance for minimum angle needed for rotation of object
    tol = MAG_TOL

    def __init__(self, center, size, **kwargs):
        """Initialises a cuboid

        Args:
            center (tuple): x,y,z
            size (tuple): x,y,z
        """
        super().__init__()

        self.center = _np.asarray(center)
        self.size = _np.asarray(size)

        self.alpha = kwargs.pop("alpha", 0.0)
        self.alpha_rad = _np.deg2rad(self.alpha)
        self.beta = kwargs.pop("beta", 0.0)
        self.beta_rad = _np.deg2rad(self.beta)
        self.gamma = kwargs.pop("gamma", 0.0)
        self.gamma_rad = _np.deg2rad(self.gamma)
        self.color = kwargs.pop("color", "white")

    def __repr__(self) -> str:
        return f"(center: {self.center}, size: {self.size} )"

    def __str__(self) -> str:
        return f"(center: {self.center}, size: {self.size} )"

    def generate_vertices(self):
        """Generates vertices of a polyhedron

        This should be implemented for each Polyhedron child class
        """
        pass


class Graphic_Cuboid(Polyhedron):
    """Generates Cuboid for plotly rendering"""

    def __init__(self, center=(0, 0, 0), size=(1, 1, 1), **kwargs):
        """Init method

        Args:
            center (tuple, optional): Cuboid center. Defaults to (0, 0, 0).
            size (tuple, optional): Size of cuboid. Defaults to (1, 1, 1).

        Kwargs:
            alpha (float): rotation wit respect to ? axis. Defaults to 0.0.
            beta (float): rotation wit respect to ? axis. Defaults to 0.0.
            gamma (float): rotation wit respect to ? axis. Defaults to 0.0.
            color (str): color. Defaults to 'w'
        """
        super().__init__(center, size, **kwargs)

        self.vertices = self.generate_vertices()

    def generate_vertices(self):
        """Generates and rotates vertices of a cuboid based on orientation angles

        Returns:
            ndarray: 3xN array of vertex coordinates (columns are x, y, z)
        """
        # Generate and rotate the vertices
        if _np.any(
            _np.fabs(
                _np.array(
                    [
                        self.alpha_rad,
                        self.beta_rad,
                        self.gamma_rad,
                    ]
                )
            )
            > Polyhedron.tol
        ):

            # _, reverse_rotation = self._generate_rotation_quaternions()
            forward_rotation = Quaternion.gen_rotation_quaternion(
                self.alpha_rad, self.beta_rad, self.gamma_rad
            )
            reverse_rotation = forward_rotation.get_conjugate()
            # Generate 3xN array for quaternion rotation

            vertex_coords = self._gen_vertices(center=(0, 0, 0), size=self.size)

            # Rotate points
            x, y, z = reverse_rotation * vertex_coords

            # Reconstruct 3xN array and add center offset
            vertex_coords = _np.vstack([x, y, z])
            vertex_coords += _np.array(self.center).reshape(-1, 1)

            # finally return the coordinates
            return vertex_coords

        # Only generate
        else:
            vertex_coords = self._gen_vertices(self.center, self.size)
            return vertex_coords

    @staticmethod
    def _gen_vertices(center=(0, 0, 0), size=(1, 1, 1)):
        """Generates coordinates for all cuboid vertices

        Args:
            center (tuple, optional): x,y,z coordinates. Defaults to (0, 0, 0)
            size (tuple, optional): scale the cuboid in x, y, z directons. Defaults to (1, 1, 1).

        Returns:
            ndarray: numpy array of shape (3, 8)
        """
        # Center of this cube is (0.5, 0.5, 0.5)
        x = _np.array([0, 0, 1, 1, 0, 0, 1, 1]) - 0.5
        y = _np.array([0, 1, 1, 0, 0, 1, 1, 0]) - 0.5
        z = _np.array([0, 0, 0, 0, 1, 1, 1, 1]) - 0.5

        vertex_coords = _np.vstack([x, y, z])

        # Scale x, y, z by the size array
        vertex_coords = _np.multiply(vertex_coords.T, size).T

        # Offset by externally set center
        vertex_coords += _np.array(center).reshape(-1, 1)

        return vertex_coords


class Graphic_Sphere(Polyhedron):
    """Generates vertices for a sphere for plotly rendering"""

    def __init__(self, center=(0, 0, 0), radius=1, **kwargs):
        """Init method

        Args:
            center (tuple, optional): [description]. Defaults to (0, 0, 0).
            radius (float, optional): [description]. Defaults to 1.

        Kwargs:
            alpha (float): rotation wit respect to ? axis. Defaults to 0.0.
            beta (float): rotation wit respect to ? axis. Defaults to 0.0.
            gamma (float): rotation wit respect to ? axis. Defaults to 0.0.
            color (str): color. Defaults to 'w'
        """
        super().__init__(center, size=(radius, radius, radius), **kwargs)
        self.radius = radius
        self.vertices = self.generate_vertices()

    def generate_vertices(self):
        """Generates and rotates vertices of a cuboid based on orientation angles

        Returns:
            ndarray: 3xN array of vertex coordinates (columns are x, y, z)
        """
        # Generate and rotate the vertices
        if _np.any(
            _np.fabs(
                _np.array(
                    [
                        self.alpha_rad,
                        self.beta_rad,
                        self.gamma_rad,
                    ]
                )
            )
            > Polyhedron.tol
        ):

            # _, reverse_rotation = self._generate_rotation_quaternions()
            forward_rotation = Quaternion.gen_rotation_quaternion(
                self.alpha_rad, self.beta_rad, self.gamma_rad
            )
            reverse_rotation = forward_rotation.get_conjugate()

            # Generate 3xN array for quaternion rotation

            vertex_coords = self._gen_vertices(center=(0, 0, 0), radius=self.radius)

            # Rotate points
            x, y, z = reverse_rotation * vertex_coords

            # Reconstruct 3xN array and add center offset
            vertex_coords = _np.vstack([x, y, z])
            vertex_coords += _np.array(self.center).reshape(-1, 1)

            # finally return the coordinates
            return vertex_coords

        # Only generate
        else:
            vertex_coords = self._gen_vertices(self.center, self.radius)
            return vertex_coords

    @staticmethod
    def _gen_vertices(center=(0, 0, 0), radius=1):
        """Generates coordinates for approximate sphere vertices

        Args:
            center (tuple, optional): x,y,z coordinates. Defaults to (0, 0, 0)
            radius (float, optional): radius of the sphere. Defaults to 1.

        Returns:
            ndarray: numpy array of shape (3, 8)
        """
        u, v = _np.mgrid[0 : 2 * PI : 20j, 0:PI:10j]
        x = radius * _np.cos(u) * _np.sin(v)
        y = radius * _np.sin(u) * _np.sin(v)
        z = radius * _np.cos(v)

        vertex_coords = _np.vstack([x.ravel(), y.ravel(), z.ravel()])

        # Offset by externally set center
        vertex_coords += _np.array(center).reshape(-1, 1)

        return vertex_coords


class Graphic_Cylinder(Polyhedron):
    """Generates vertices for a Cylinder for plotly rendering"""

    def __init__(self, center=(0, 0, 0), radius=1, length=1, **kwargs):
        """Init method

        Args:
            center (tuple, optional): [description]. Defaults to (0, 0, 0).
            radius (float, optional): [description]. Defaults to 1.
            length (float, optional): [description]. Defaults to 1.

        Kwargs:
            alpha (float): rotation wit respect to ? axis. Defaults to 0.0.
            beta (float): rotation wit respect to ? axis. Defaults to 0.0.
            gamma (float): rotation wit respect to ? axis. Defaults to 0.0.
            color (str): color. Defaults to 'w'
        """
        super().__init__(center, size=(radius, radius, length), **kwargs)
        self.radius = radius
        self.length = length
        self.vertices = self.generate_vertices()

    def generate_vertices(self):
        """Generates and rotates vertices of a cuboid based on orientation angles

        Returns:
            ndarray: 3xN array of vertex coordinates (columns are x, y, z)
        """
        # Generate and rotate the vertices
        if _np.any(
            _np.fabs(
                _np.array(
                    [
                        self.alpha_rad,
                        self.beta_rad,
                        self.gamma_rad,
                    ]
                )
            )
            > Polyhedron.tol
        ):

            # _, reverse_rotation = self._generate_rotation_quaternions()
            forward_rotation = Quaternion.gen_rotation_quaternion(
                self.alpha_rad, self.beta_rad, self.gamma_rad
            )
            reverse_rotation = forward_rotation.get_conjugate()

            # Generate 3xN array for quaternion rotation

            vertex_coords = self._gen_vertices(
                center=(0, 0, 0), radius=self.radius, length=self.length
            )

            # Rotate points
            x, y, z = reverse_rotation * vertex_coords

            # Reconstruct 3xN array and add center offset
            vertex_coords = _np.vstack([x, y, z])
            vertex_coords += _np.array(self.center).reshape(-1, 1)

            # finally return the coordinates
            return vertex_coords

        # Only generate
        else:
            vertex_coords = self._gen_vertices(self.center, self.radius, self.length)
            return vertex_coords

    @staticmethod
    def _gen_vertices(center=(0, 0, 0), radius=1, length=1):
        """Generates coordinates for approximate cylinder vertices

        Args:
            center (tuple, optional): x,y,z coordinates. Defaults to (0, 0, 0)
            radius (float, optional): radius of the cylinder. Defaults to 1.
            length (float, optional): length of the cylinder. Defaults to 1.


        Returns:
            ndarray: numpy array of shape (3, 8)
        """
        rho, z = _np.mgrid[0 : 2 * PI : 40j, -length / 2 : length / 2 : 2j]
        x = radius * _np.cos(rho)
        y = radius * _np.sin(rho)

        vertex_coords = _np.vstack([x.ravel(), y.ravel(), z.ravel()])

        # Offset by externally set center
        vertex_coords += _np.array(center).reshape(-1, 1)
        return vertex_coords


class Graphic_Mesh(Polyhedron):
    """Generates Mesh from STL file for plotly rendering"""

    def __init__(self, mesh_vectors, **kwargs):
        """Init Method

        Args:
            mesh_vectors (ndarray): array of mesh vectors

        Kwargs:
            color (str): magnet color. Defaults to 'w'.
        """
        self.color = kwargs.pop("color", "white")
        self.mesh_vectors = mesh_vectors

    def generate_vertices(self):
        """Generates vertices from STL file

        Returns:
            dict: Dictionary of rendering properties for plotly
        """
        # return super().generate_vertices()
        p, q, r = self.mesh_vectors.shape  # (p, 3, 3)

        # the array stl_mesh.vectors.reshape(p*q, r) can contain multiple copies of the same vertex;
        # extract unique vertices from all mesh triangles
        vertices, ixr = _np.unique(
            self.mesh_vectors.reshape(p * q, r), return_inverse=True, axis=0
        )
        I = _np.take(ixr, [3 * k for k in range(p)])
        J = _np.take(ixr, [3 * k + 1 for k in range(p)])
        K = _np.take(ixr, [3 * k + 2 for k in range(p)])
        x, y, z = vertices.T
        trace = _go.Mesh3d(x=x, y=y, z=z, i=I, j=J, k=K, color=self.color)

        # optional parameters to make it look nicer
        trace.update(
            flatshading=True, lighting_facenormalsepsilon=0, lighting_ambient=0.7
        )
        return trace


def reset_polyhedra():
    """Returns a list of all instantiated polyhedra."""

    polyhedra_classes = [
        Polyhedron,
        Graphic_Cuboid,
        Graphic_Sphere,
        Graphic_Cylinder,
    ]
    for cls in polyhedra_classes:
        cls.reset()


def list_polyhedra():
    """Returns a list of all instantiated polyhedra.

    Assumes that the child class registries have not been modified outside of
    using `pymagnet.reset_magnets()`.
    """
    return Polyhedron.print_instances()


def _draw_surface_slice(
    points,
    field,
    colorscale="viridis",
    opacity=1.0,
    cmin=0.0,
    cmax=1.0,
    showscale=True,
):
    """Generates plotly surface object for plotting

    Args:
        points (Point_Array3): coordinates
        field (Field3): Magnetic field vectors
        colorscale (str, optional): colormap. Defaults to "viridis".
        opacity (float, optional): Opacity of slice. Defaults to 1.0.
        cmin (float, optional): minimum color scale value. Defaults to 0.0.
        cmax (float, optional): maximum color scale value. Defaults to 1.0.
        showscale (bool, optional): show colorbar. Defaults to True.

    Returns:
        dict: plotly surface data structure for plotting
    """
    return _go.Surface(
        x=points.x,
        y=points.y,
        z=points.z,
        surfacecolor=field.n,
        colorscale=colorscale,
        opacity=opacity,
        showscale=showscale,
        cmin=cmin,
        cmax=cmax,
        colorbar=dict(title="|B| (" + field.unit + ")"),
    )


def _draw_mesh(vertices, magnet_opacity, magnet_color):
    """Generates mesh of magnet for drawing in plotly plot

    Args:
        vertices (ndarray): array of vertices
        magnet_opacity (float): opacity of object

    Returns:
        Mesh3d object: plotly graphics object (Mesh3d)
    """
    return _go.Mesh3d(
        x=vertices[0],
        y=vertices[1],
        z=vertices[2],
        alphahull=0,
        opacity=magnet_opacity,
        color=magnet_color,
        # vertexcolor=(0, 0, 0),
        flatshading=True,
        hoverinfo="skip",
        name="y",
        showscale=False,
    )


def _draw_cones(points, field, NA=10, cone_opacity=1.0):
    """Generates cones object for drawing vector field in 3D in plotly.
    Arrows are white and normalised.

    Args:
        points (Point_Array3): coordinates
        field (Field3): Magnetic field vector
        NA (int, optional): Number of cones per axis. Defaults to 10.
        cone_opacity (float, optional): opacity of cones. Defaults to 1.0.

    Returns:
        dict: cone data structure for plotting
    """
    x = points.x.reshape(field.x.shape)
    y = points.y.reshape(field.y.shape)
    z = points.z.reshape(field.z.shape)
    maxB = field.n[::NA, ::NA].ravel()
    cscale = [[0, "rgb(255,255,255)"], [1, "rgb(255,255,255)"]]
    data_object = _go.Cone(
        x=x[::NA, ::NA].ravel(),
        y=y[::NA, ::NA].ravel(),
        z=z[::NA, ::NA].ravel(),
        u=field.x[::NA, ::NA].ravel() / maxB,
        v=field.y[::NA, ::NA].ravel() / maxB,
        w=field.z[::NA, ::NA].ravel() / maxB,
        sizemode="scaled",
        colorscale=cscale,
        showscale=False,
        opacity=cone_opacity,
    )
    return data_object


def _generate_all_meshes(magnet_opacity=1.0, **kwargs):
    """Generates polyhedra of all magnets for rendering in 3D

    Args:
        magnet_opacity (float, optional): Opacity of magnets. Defaults to 1.0.

    Returns:
        dict: plotly data structure for rendering magnets
    """
    color = kwargs.pop("color", "white")
    data_objects = []
    for magnet in Magnet3D.instances:
        if issubclass(magnet.__class__, Mesh):
            mesh_data = Graphic_Mesh(magnet.mesh_vectors, color=color)
            data_objects.append(mesh_data.generate_vertices())

        else:
            if issubclass(magnet.__class__, Prism):
                polyhed = Graphic_Cuboid(
                    center=magnet.center,
                    size=magnet.get_size(),
                    alpha=magnet.alpha,
                    beta=magnet.beta,
                    gamma=magnet.gamma,
                    color=color,
                )
            elif issubclass(magnet.__class__, Cylinder):
                polyhed = Graphic_Cylinder(
                    center=magnet.center,
                    radius=magnet.radius,
                    length=magnet.length,
                    alpha=magnet.alpha,
                    beta=magnet.beta,
                    gamma=magnet.gamma,
                    color=color,
                )

            elif issubclass(magnet.__class__, Sphere):
                polyhed = Graphic_Sphere(
                    center=magnet.center,
                    radius=magnet.radius,
                    alpha=magnet.alpha,
                    beta=magnet.beta,
                    gamma=magnet.gamma,
                    color=color,
                )

            data_objects.append(
                _draw_mesh(polyhed.vertices, magnet_opacity, polyhed.color)
            )

    return data_objects


def _generate_volume_data(points, field, **kwargs):
    """Generates volume data plot for plotly

    Args:
        points (Point_Array3): coordinates
        field (Field3): Magnetic field vector

    Returns:
        dict: plotly volume data structure
    """

    cmin = kwargs.pop("cmin", 0.0)
    cmax = kwargs.pop("cmax", 0.5)
    colorscale = kwargs.pop("colorscale", "viridis")

    opacityscale = kwargs.pop("opacityscale", None)

    caps = kwargs.pop("no_caps", False)
    if caps:
        caps = dict(x_show=False, y_show=False, z_show=False)
    else:
        caps = dict(x_show=True, y_show=True, z_show=True)

    if type(opacityscale) is str:
        if opacityscale.lower() == "normal":
            opacityscale = [
                [cmin, 0],
                [(cmax - cmin) / 4, 0.5],
                [0.2, 0],
                [cmax, 1],
            ]
        elif opacityscale.lower() == "invert":
            opacityscale = [
                [cmin, 1],
                [(cmax - cmin) * 3 / 4, 0.5],
                [0.2, 0],
                [cmax, 0],
            ]

    return _go.Volume(
        x=points.x.flatten(),
        y=points.y.flatten(),
        z=points.z.flatten(),
        value=field.n.flatten(),
        colorscale=colorscale,
        cmin=cmin,
        cmax=cmax,
        isomin=cmin,
        isomax=cmax,
        # opacity needs to be small to see through all surfaces
        opacity=kwargs.pop("opacity", 0.1),
        opacityscale=opacityscale,
        surface_count=kwargs.pop("num_levels", 10),
        showscale=True,
        caps=caps,
        colorbar=dict(title="|B| (" + field.unit + ")"),
    )


def plot_magnet(unit="mm", **kwargs):
    """Renders magnets

    Args:
        unit (str, optional): unit scale. Defaults to 'mm'.

    Returns:
        fig: reference to figure
    """

    reset_polyhedra()

    magnet_opacity = kwargs.pop("magnet_opacity", 1.0)
    data_objects = []

    data_objects.extend(_generate_all_meshes(magnet_opacity=magnet_opacity))

    fig = _go.Figure(data=data_objects)

    fig.update_layout(
        scene=dict(
            xaxis_title="x (" + unit + ")",
            yaxis_title="y (" + unit + ")",
            zaxis_title="z (" + unit + ")",
        ),
        width=700,
        margin=dict(r=20, b=10, l=10, t=10),
    )
    fig.update_layout(scene_aspectmode="data")
    fig.show()
    return fig


def slice_plot(data_dict, **kwargs):
    """Plots magnetic field slices.
    A convenience function.

    Returns:
        tuple: fig (reference to figure), data_objects (plotly dict)
    """

    reset_polyhedra()

    opacity = kwargs.pop("opacity", 0.8)
    magnet_opacity = kwargs.pop("magnet_opacity", 1.0)
    cone_opacity = kwargs.pop("cone_opacity", 1.0)
    # planes = kwargs.pop("planes", ["xy", "xz", "yz"])

    cmin = kwargs.pop("cmin", 0)
    cmax = kwargs.pop("cmax", 0.5)
    colorscale = kwargs.pop("colorscale", "viridis")
    num_arrows = kwargs.pop("num_arrows", None)
    # vector_plot = kwargs.pop("vector_plot", True)

    data_objects = []

    show_magnets = kwargs.pop("show_magnets", True)

    if show_magnets:
        data_objects.extend(_generate_all_meshes(magnet_opacity=magnet_opacity))

    for plane in data_dict.keys():
        points = data_dict[plane]["points"]
        field = data_dict[plane]["field"]

        data_objects.append(
            _draw_surface_slice(
                points,
                field,
                colorscale,
                opacity=opacity,
                cmin=cmin,
                cmax=cmax,
                showscale=True,
            )
        )
        if num_arrows is not None:
            num_points = field.x.shape[0]

            NA = num_points // num_arrows

            if NA < 1:
                NA = 1
            data_objects.append(
                _draw_cones(points, field, NA=NA, cone_opacity=cone_opacity)
            )

    fig = _go.Figure(data=data_objects)

    fig.update_layout(
        scene=dict(
            xaxis_title="x (" + points.unit + ")",
            yaxis_title="y (" + points.unit + ")",
            zaxis_title="z (" + points.unit + ")",
        ),
        width=700,
        margin=dict(r=20, b=10, l=10, t=10),
    )
    fig.update_layout(scene_aspectmode="data")
    fig.show()
    return fig, data_objects


def slice_quickplot(**kwargs):
    """Calculates and plots magnetic field slices.
    A convenience function.

    Returns:
        tuple: fig (reference to figure), cache (cached data for each plane with potential keys: 'xy', 'xz', 'yz'), data_objects (plotly dict)
    """

    reset_polyhedra()

    max1 = kwargs.pop("max1", 30)
    max2 = kwargs.pop("max2", 30)
    min1 = kwargs.pop("min1", -1 * max1)
    min2 = kwargs.pop("min2", -1 * max2)

    slice_value = kwargs.pop("slice_value", 0.0)
    unit = kwargs.pop("unit", "mm")

    opacity = kwargs.pop("opacity", 0.8)
    magnet_opacity = kwargs.pop("magnet_opacity", 1.0)
    cone_opacity = kwargs.pop("cone_opacity", 1.0)
    planes = kwargs.pop("planes", ["xy", "xz", "yz"])

    num_arrows = kwargs.pop("num_arrows", None)
    num_points = kwargs.pop("num_points", 100)

    if num_arrows is not None:
        NA = num_points // num_arrows
        if NA < 1:
            NA = 1

    cmin = kwargs.pop("cmin", 0)
    cmax = kwargs.pop("cmax", 0.5)
    colorscale = kwargs.pop("colorscale", "viridis")

    data_objects = []
    cache = {}

    show_magnets = kwargs.pop("show_magnets", True)

    if show_magnets:
        data_objects.extend(_generate_all_meshes(magnet_opacity=magnet_opacity))

    for plane in planes:
        points = slice3D(
            plane=plane,
            max1=max1,
            min1=min1,
            max2=max2,
            min2=min2,
            slice_value=slice_value,
            unit=unit,
            num_points=num_points,
        )
        field = get_field_3D(points)

        cache[plane] = {"points": points, "field": field}

        data_objects.append(
            _draw_surface_slice(
                points,
                field,
                colorscale,
                opacity=opacity,
                cmin=cmin,
                cmax=cmax,
                showscale=True,
            )
        )
        if num_arrows is not None:
            data_objects.append(
                _draw_cones(points, field, NA=NA, cone_opacity=cone_opacity)
            )

    fig = _go.Figure(data=data_objects)

    fig.update_layout(
        scene=dict(
            xaxis_title="x (" + points.unit + ")",
            yaxis_title="y (" + points.unit + ")",
            zaxis_title="z (" + points.unit + ")",
        ),
        width=700,
        margin=dict(r=20, b=10, l=10, t=10),
    )
    fig.update_layout(scene_aspectmode="data")
    fig.show()
    return fig, cache, data_objects


def volume_plot(points, field, **kwargs):
    """Plots magnetic field volume.

    Args:
        points (Point_Array3): coordinates
        field (Field3): Magnetic field vector

    Returns:
        tuple: fig (reference to figure), data_objects (plotly dict)
    """

    reset_polyhedra()

    opacity = kwargs.pop("opacity", 1)
    opacityscale = kwargs.pop("opacityscale", None)
    magnet_opacity = kwargs.pop("magnet_opacity", 1.0)
    cone_opacity = kwargs.pop("cone_opacity", 1.0)

    num_arrows = kwargs.pop("num_arrows", None)

    cmin = kwargs.pop("cmin", 0)
    cmax = kwargs.pop("cmax", 0.5)
    num_levels = kwargs.pop("num_levels", 5)

    show_magnets = kwargs.pop("show_magnets", True)

    num_points = len(points.x)

    data_objects = []

    if show_magnets:
        data_objects.extend(_generate_all_meshes(magnet_opacity=magnet_opacity))

    colorscale = kwargs.pop("colorscale", "viridis")

    #     kernel_size = 1
    #     kernel = np.ones([kernel_size, kernel_size, kernel_size]) / kernel_size
    #     B.n = ndimage.convolve(B.n, kernel)

    data_objects.append(
        _generate_volume_data(
            points,
            field,
            cmim=cmin,
            cmax=cmax,
            opacity=opacity,
            colorscale=colorscale,
            num_levels=num_levels,
            opacityscale=opacityscale,
        )
    )

    if num_arrows is not None:
        NA = num_points // num_arrows
        if NA < 1:
            NA = 1
        data_objects.append(
            _draw_cones(points, field, NA=NA, cone_opacity=cone_opacity)
        )

    fig = _go.Figure(data=data_objects)

    fig.update_layout(
        scene=dict(
            xaxis_title="x (" + points.unit + ")",
            yaxis_title="y (" + points.unit + ")",
            zaxis_title="z (" + points.unit + ")",
        ),
        width=700,
        margin=dict(r=20, b=10, l=10, t=10),
    )
    fig.update_layout(scene_aspectmode="data")
    fig.show()

    return fig, data_objects


def volume_quickplot(**kwargs):
    """Calculates and plots magnetic field slices.
    A convenience function.

    Kwargs:
        num_points (int): = kwargs.pop("num_points", 30)
        unit (str): = kwargs.pop("unit", "mm")
        xmax (float): Maximum x value. Defaults to 30.0.
        ymax (float): Maximum y value. Defaults to 30.0.
        zmax (float): Maximum z value. Defaults to 30.0.
        xmin (float): Minimum x value. Defaults to -xmax
        ymin (float): Minimum y value. Defaults to -ymax
        zmin (float): Minimum z value. Defaults to -zmax

    Returns:
        tuple: fig (reference to figure), cache (cached data dict), data_objects (plotly dict)
    """

    num_points = kwargs.pop("num_points", 30)

    unit = kwargs.pop("unit", "mm")

    xmax = kwargs.pop("xmax", 30)
    ymax = kwargs.pop("ymax", 30)
    zmax = kwargs.pop("zmax", 30)

    xmin = kwargs.pop("xmin", -1 * xmax)
    ymin = kwargs.pop("ymin", -1 * ymax)
    zmin = kwargs.pop("zmin", -1 * zmax)

    points = grid3D(
        xmax,
        ymax,
        zmax,
        num_points=num_points,
        xmin=xmin,
        ymin=ymin,
        zmin=zmin,
        unit=unit,
    )
    field = get_field_3D(points)

    fig, data_objects = volume_plot(points, field, num_points=num_points, **kwargs)
    cache = {"points": points, "field": field}

    return fig, cache, data_objects


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""Contains functions needed to rotate and translate a triangle to lie in the xz plane
and to divide it into two right angled triangles
"""
from ._quaternion import Quaternion
from .global_const import PI, ALIGN_CUTOFF

import numpy as _np
from numba import jit

__all__ = ["signed_area", "norm_plane", "rotate_points", "altitude"]


@jit
def signed_area(triangle):
    """Calculates signed area of a triangle. Area area < 0 for clockwise ordering.
    Assumes the triangle is in the xz plane (i.e. with the normal parallel to y).

    Args:
        triangle (ndarray): 3x3 array of vertices

    Returns:
        float: signed area
    """

    j = 1
    NP = 3
    area = 0.0

    for i in range(NP):
        j = j % NP

        area += (triangle[j][0] - triangle[i][0]) * (triangle[j][2] + triangle[i][2])
        j += 1

    # check winding order of polygon, area < 0 for clockwise ordering of points
    area /= 2.0

    return area


def norm_plane(vec):
    """Calculates the normal to a triangular plane

    Args:
        vec (ndarray/list/tuple): (N,1) array

    Returns:
        ndarray: normal vector (N,)
    """
    norm = _np.cross(vec[1] - vec[0], vec[2] - vec[0])
    norm = norm / _np.linalg.norm(norm)
    return norm


def rotate_points(points, rotation_quaternion):
    """Rotates a set of points

    Args:
        points ([type]): [description]
        rotation_quaternion ([type]): [description]

    Returns:
        [type]: [description]
    """
    x_rot, y_rot, z_rot = rotation_quaternion * points.T
    rotate_points = _np.vstack([x_rot, y_rot, z_rot]).T

    return rotate_points


def altitude(a, b, c):
    """Gets altitude to side `a` of a triangle.

    Args:
        a (float): longest side
        b (float): triangle side
        c (float): triangle side

    Returns:
        float: altitude to side `a`
    """
    s = (a + b + c) / 2
    return 2 * _np.sqrt(s * (s - a) * (s - b) * (s - c)) / a


def _largest_side_RA(triangle):
    """Determines largest side of triangle

    Args:
        triangle (ndarray): triangle vertices

    Returns:
        tuple: longest_side (int), length (float), right_angle (bool)
    """
    A = _np.linalg.norm([triangle[1] - triangle[0]])
    B = _np.linalg.norm([triangle[2] - triangle[1]])
    C = _np.linalg.norm([triangle[2] - triangle[0]])
    sides = _np.array([[0, 1, 2], [A, B, C]]).T
    sorted_sides = sides[sides[:, 1].argsort()][::-1]
    longest_side = int(sorted_sides[0, 0])

    # Get sizes of two right angled triangles

    # Get Altitude to longest side
    alt_side = altitude(sorted_sides[0, 1], sorted_sides[1, 1], sorted_sides[2, 1])
    left_side = sides[(longest_side + 1) % 3, 1]
    right_side = sides[(longest_side - 1) % 3, 1]

    p = _np.sqrt(left_side ** 2 - alt_side ** 2)
    q = _np.sqrt(right_side ** 2 - alt_side ** 2)

    RA_triangle1 = _np.array([p, alt_side])
    RA_triangle2 = _np.array([q, alt_side])

    return longest_side, RA_triangle1, RA_triangle2


def check_sign(vector_1, vector_2):
    """Returns true if the signs of all elements of two arrays are the same

    Args:
        vector_1 (ndarray): input array 2
        vector_2 (ndarray): input array 2

    Returns:
        boolean: True if elements in two arrays have the same sign
    """
    sign_comp_1 = _np.fabs(vector_1 + vector_2)
    sign_comp_2 = _np.fabs(vector_1) + _np.fabs(vector_2)

    return _np.allclose(sign_comp_1, sign_comp_2, atol=1e-6)


def return_axis_vector(triangle, longest_side):
    """Returns vector collinear to the longest side of a triangle

    Args:
        triangle (ndarray): triangle vertices
        longest_side (int): longest side index

    Returns:
        ndarray: vector corresponding to longest side
    """
    vector_A = triangle[1] - triangle[0]
    vector_A = vector_A / _np.linalg.norm(vector_A)

    vector_B = triangle[2] - triangle[1]
    vector_B = vector_B / _np.linalg.norm(vector_B)
    vector_C = triangle[2] - triangle[0]
    vector_C = vector_C / _np.linalg.norm(vector_C)

    vec_dict = {
        0: vector_A,
        1: vector_B,
        2: vector_C,
    }

    vec = vec_dict[longest_side]

    return vec


def return_z_vector(triangle, longest_side):
    """Returns altitude vector from longest side

    Args:
        triangle (ndarray): triangle vertices
        longest_side (int): longest side index

    Returns:
        ndarray: altitude vector
    """
    new_vertex_dict = {
        0: [triangle[2, 0], triangle[0, 1], triangle[0, 2]],
        1: [triangle[0, 0], triangle[0, 1], triangle[1, 2]],
        2: [triangle[1, 0], triangle[2, 1], triangle[2, 2]],
    }
    new_vertex = _np.array(new_vertex_dict[longest_side])
    vec_z_dict = {
        0: triangle[2] - new_vertex,
        1: triangle[0] - new_vertex,
        2: triangle[1] - new_vertex,
    }
    vec_z = vec_z_dict[longest_side]
    vec_z = vec_z / _np.linalg.norm(vec_z)

    return vec_z


def align_triangle_to_y(triangle, rot_axis, norm_vec):
    """Rotates and translates a triangle in lie in the xz plane

    Args:
        triangle (ndarray): vertices of a triangle
        rot_axis (ndarray): axis about which to rotate triangle
        norm_vec (ndarray): normal to triangle

    Returns:
        tuple: aligned_triangle (ndarray, rotated triangle), first_rotation (quaternion, align to y-axis)
    """
    y_axis = _np.array([0, 1, 0])

    if _np.linalg.norm(rot_axis) < ALIGN_CUTOFF:

        # Check if parallel or anti-parallel
        if check_sign(y_axis, norm_vec):
            # Parallel
            first_rotation = Quaternion()
            aligned_triangle = triangle

        else:
            # Anti-parallel
            first_rotation = Quaternion.q_angle_from_axis(PI, y_axis)
            aligned_triangle = rotate_points(triangle, first_rotation)

    else:
        angle = -_np.arccos(_np.dot(y_axis, norm_vec))
        first_rotation = Quaternion.q_angle_from_axis(angle, rot_axis)
        aligned_triangle = rotate_points(triangle, first_rotation)
    return aligned_triangle, first_rotation


def align_triangle_xz(triangle, longest_side):
    """Returns quaternions needed to rotate a triangle lying in the xz plane to be
    aligned with its longest side along x and altitude along z

    Args:
        triangle (ndarray): vertices of triangle
        longest_side (int): index of the longest side of triangle

    Returns:
        tuple: second_rotation (quaternion, align to x-axis), third_rotation (quaternion, align to z-axis):
    """
    x_axis = _np.array([1, 0, 0])
    y_axis = _np.array([0, 1, 0])
    z_axis = _np.array([0, 0, 1])

    # side_list = [0, 1, 2]
    # side_list.pop(longest_side)

    vec_x = return_axis_vector(triangle, longest_side)
    rot_axis = _np.cross(x_axis, vec_x)

    # Check aligment of base of triangle with x-axis
    if _np.linalg.norm(rot_axis) < ALIGN_CUTOFF:

        # Check if parallel or anti-parallel
        if check_sign(x_axis, vec_x):
            # Parallel
            second_rotation = Quaternion()
            tri_x = triangle
        else:
            # Anti-parallel
            second_rotation = Quaternion.q_angle_from_axis(PI, y_axis)
            tri_x = rotate_points(triangle, second_rotation)

    else:
        angle = -_np.arccos(_np.dot(x_axis, vec_x))
        second_rotation = Quaternion.q_angle_from_axis(angle, rot_axis)
        tri_x = rotate_points(triangle, second_rotation)

    vec_z = return_z_vector(tri_x, longest_side)
    rot_axis = _np.cross(z_axis, vec_z)

    # Check aligment of triangle altitude with z-axis
    if _np.all(_np.fabs([rot_axis]) < ALIGN_CUTOFF):

        # Check if parallel anti-parallel
        if check_sign(z_axis, vec_z):
            # Parallel
            third_rotation = Quaternion()
        else:
            # Anti-parallel
            third_rotation = Quaternion.q_angle_from_axis(PI, y_axis)

    else:
        angle = -_np.arccos(_np.dot(z_axis, vec_z))
        third_rotation = Quaternion.q_angle_from_axis(angle, rot_axis)

    return second_rotation, third_rotation


def _rotate_triangle(triangle, Jr):
    """Gets rotation angles needed for transformation of coordinates from
    global frame to a local frame

    Args:
        triangle (ndarray): vertices of a triangular plane

    Returns:
        tuple: rotated_triangle (ndarray), right_angled (Boolean), total_rotation (Quaternion)
    """

    loc_triangle = triangle - triangle[0]

    y_axis = _np.array([0, 1, 0])

    longest_side, RA_triangle1, RA_triangle2 = _largest_side_RA(loc_triangle)

    # align norm to y axis
    norm_vec_y = norm_plane(loc_triangle)
    rot_axis = _np.cross(y_axis, norm_vec_y)

    aligned_triangle, first_rotation = align_triangle_to_y(
        loc_triangle, rot_axis, norm_vec_y
    )

    second_rotation, third_rotation = align_triangle_xz(aligned_triangle, longest_side)
    total_rotation = third_rotation * second_rotation * first_rotation

    rotated_triangle = rotate_points(triangle, total_rotation)

    origin_vertex = _np.argwhere(
        rotated_triangle[:, 0] == rotated_triangle[:, 0].min()
    ).ravel()[0]

    offset = rotated_triangle[origin_vertex]

    return total_rotation, rotated_triangle, offset, RA_triangle1, RA_triangle2


__all__ = ["calc_force_cylinder"]
import numpy as _np
from ..utils.global_const import MU0, PI
from ..utils._conversions import get_unit_value_meter
from ..utils._vector_structs import Point_Array3
from ..utils._quaternion import Quaternion


from ._prism_force import _calc_field_face


def _get_ranges_cylinder(active_magnet):
    """Gets min and max x,y,z values for cylinder magnet faces.

    This is needed to generate a grid of points on each surface for calculating
    the force and torque on the magnet due to the other magnets.

    Args:
        active_magnet (Cylinder): Target magnet

    Returns:
        tuple: xlim (tuple), ylim (tuple), zlim (tuple), areas (ndarray) - areas of each face of the cuboid
    """
    radius, length = active_magnet.get_size()
    xlim = (-radius, radius)
    ylim = (-radius, radius)
    zlim = (-length / 2, length / 2)

    areas = _np.array([1, 1]) * PI * radius ** 2
    return xlim, ylim, zlim, areas


def _gen_cylinder_grid(active_magnet, xlim, ylim, num_segments=10):
    """Generates a 2D grid of N**2 points for a given pair of x and y limits.

    Args:
        active_magnet (Cylinder): Target magnet.
        xlim (tuple, optional): Min and max x values. Defaults to (0, 10).
        ylim (tuple, optional): Min and max y values. Defaults to (0, 10).
        num_rectangles (int, optional): Number of points to generate in each direction. Defaults to 10.

    Returns:
        ndarray: (num_rectangles**2, 3) array of points
    """
    NS = num_segments * 1j
    radius, _ = active_magnet.get_size()
    xc, yc, _ = active_magnet.get_center()
    x, y = _np.mgrid[xlim[0] : xlim[1] : NS, ylim[0] : ylim[1] : NS]
    index = (x - xc) ** 2 + (y - yc) ** 2 < radius ** 2
    x = x[index]
    y = y[index]

    return x, y


def _gen_cylinder_face_grid(
    active_magnet, xlim, ylim, zlim, num_segments=20, unit="mm", reverse_rotation=None
):
    """Returns grids of points for the two circular faces of a cylinder

    Args:
        xlim (tuple): min and max x values (float)
        ylim (tuple): min and max y values (float)
        zlim (tuple): min and max z values (float)
        num_segments (int, optional): Number of grid points to generate (10x10). Defaults to 20.
        unit (str, optional): Length scale. Defaults to "mm".

    Returns:
        tuple: points_lower, points_upper both (num_rectangles**2, 3) ndarrays
    """
    xc, yc, zc = active_magnet.get_center()

    x, y = _gen_cylinder_grid(active_magnet, xlim, ylim, num_segments=num_segments)
    z = _np.tile([zlim[0]], x.shape)

    if reverse_rotation is not None:
        pos_vec = Quaternion._prepare_vector(x, y, z)
        x_rot, y_rot, z_rot = reverse_rotation * pos_vec
        points_lower = Point_Array3(
            x_rot.ravel() + xc, y_rot.ravel() + yc, z_rot.ravel() + zc, unit=unit
        )

        z = _np.tile([zlim[1]], x.shape)
        pos_vec = Quaternion._prepare_vector(x, y, z)
        x_rot, y_rot, z_rot = reverse_rotation * pos_vec
        points_upper = Point_Array3(
            x_rot.ravel() + xc, y_rot.ravel() + yc, z_rot.ravel() + zc, unit=unit
        )

    else:
        points_lower = Point_Array3(
            x.ravel() + xc, y.ravel() + yc, z.ravel() + zc, unit=unit
        )
        z = _np.tile([zlim[1]], x.shape)
        points_upper = Point_Array3(
            x.ravel() + xc, y.ravel() + yc, z.ravel() + zc, unit=unit
        )

    return points_lower, points_upper


# def _calc_field_face(active_magnet, points):
#     field = _allocate_field_array3(points.x, points.y, points.z)
#     for magnet in Magnet3D.instances:
#         if magnet is not active_magnet:
#             #             print(magnet, active_magnet)
#             Bx, By, Bz = magnet.get_field(points.x, points.y, points.z)
#             field.x += Bx.reshape(field.x.shape)
#             field.y += By.reshape(field.y.shape)
#             field.z += Bz.reshape(field.z.shape)
#     xc, yc, zc = active_magnet.get_center()
#     field.x[~_np.isfinite(field.x)] = 0.0
#     field.y[~_np.isfinite(field.y)] = 0.0
#     field.z[~_np.isfinite(field.z)] = 0.0
#     total_field = _np.array((_np.sum(field.x), _np.sum(field.y), _np.sum(field.z)))
#
#     torques = _np.cross(
#         _np.array(
#             [points.x.ravel() - xc, points.y.ravel() - yc, points.z.ravel() - zc]
#         ).T,
#         _np.narray([field.x.ravel(), field.y.ravel(), field.z.ravel()]).T,
#     )
#     total_torque = _np.sum(torques, axis=0)
#
#     return total_field, total_torque


def calc_force_cylinder(active_magnet, num_segments=20, unit="mm"):
    """Calculates the total force on a cylinder magnet due to all other instantiated magnets

    Args:
        active_magnet (Cylinder): Target Magnet
        num_rectangles (int, optional): Number of grid points per face. Defaults to 20.
        unit (str, optional): Length scale. Defaults to "mm".

    Returns:
        total_field (float), total_torque (float)
    """
    Jvec = active_magnet.get_Jr()

    if _np.any(
        _np.fabs(
            _np.array(
                [
                    active_magnet.alpha_rad,
                    active_magnet.beta_rad,
                    active_magnet.gamma_rad,
                ]
            )
        )
        > active_magnet.tol
    ):

        _, reverse_rotation = active_magnet._generate_rotation_quaternions()
        Jrot = reverse_rotation * Jvec
    else:
        reverse_rotation = None
        Jrot = Jvec

    xlim, ylim, zlim, areas = _get_ranges_cylinder(active_magnet)

    points_lower, points_upper = _gen_cylinder_face_grid(
        active_magnet,
        xlim,
        ylim,
        zlim,
        num_segments=num_segments,
        unit=unit,
        reverse_rotation=reverse_rotation,
    )
    num_points = _np.size(points_lower.x)

    force = _np.zeros(3)
    torque = _np.zeros(3)

    total_field, total_torque = _calc_field_face(active_magnet, points_lower)

    force += total_field * -Jrot[2] * areas[0] / num_points
    torque += total_torque * -Jrot[2] * areas[0] / num_points

    total_field, total_torque = _calc_field_face(active_magnet, points_upper)
    force += total_field * Jrot[2] * areas[1] / num_points
    torque += total_torque * Jrot[2] * areas[1] / num_points

    scaling_factor = get_unit_value_meter(points_lower.get_unit())
    force /= MU0 / scaling_factor / scaling_factor
    torque /= MU0 / scaling_factor / scaling_factor / scaling_factor
    return force, torque


from ._magnet3D import Magnet3D
from ..utils._quaternion import Quaternion
from ..utils.global_const import MAG_TOL, PI, FP_CUTOFF, ALIGN_CUTOFF
from ..utils._trigonometry3D import norm_plane, _rotate_triangle

from stl import mesh
import numpy as _np
from numba import jit, vectorize, float64
from math import sqrt, log, fabs, atan2


class Mesh(Magnet3D):
    """Mesh Magnet Class."""

    mag_type = "Mesh"

    def __init__(
        self,
        filename,
        Jr=1.0,  # local magnetisation
        **kwargs,
    ):
        """Init Method

        Args:
            filename (string): path to stl file to be imported
            Jr (float, optional): Signed remnant magnetisation. Defaults to 1.0.

        Kwargs:
            phi (float):
            theta (float):
            mesh_scale (float): scaling factor if mesh needs to be resized. Defaults to 1.0
        """
        super().__init__(Jr, **kwargs)

        self.phi = kwargs.pop("theta", 90.0)
        self.phi_rad = _np.deg2rad(self.phi)
        self.theta = kwargs.pop("phi", 0.0)
        self.theta_rad = _np.deg2rad(self.theta)

        self.mesh_scale = kwargs.pop("mesh_scale", 1.0)
        self._filename = filename

        (
            self.mesh_vectors,
            self.mesh_normals,
            self.volume,
            self.centroid,
        ) = self._import_mesh()

        self.Jx = _np.around(
            Jr * _np.cos(self.phi_rad) * _np.sin(self.theta_rad), decimals=6
        )
        self.Jy = _np.around(
            Jr * _np.sin(self.phi_rad) * _np.sin(self.theta_rad), decimals=6
        )
        self.Jz = _np.around(Jr * _np.cos(self.theta_rad), decimals=6)
        self.tol = MAG_TOL  # sufficient for 0.01 degree accuracy
        self.J = _np.array([self.Jx, self.Jy, self.Jz])

        # FIXME: Sort out rotation of magnetisation with rotation of mesh
        # if _np.any(
        #     _np.fabs([self.alpha_rad, self.beta_rad, self.gamma_rad]) > ALIGN_CUTOFF
        # ):
        #     mag_rotation = Quaternion.gen_rotation_quaternion(
        #         self.alpha_rad, self.beta_rad, self.gamma_rad
        #     )
        #     Jrot = mag_rotation * self.J
        #     self.Jx = Jrot[0]
        #     self.Jy = Jrot[1]
        #     self.Jz = Jrot[2]
        #     self.J = _np.array([self.Jx, self.Jy, self.Jz])

        self.Jnorm = _np.dot(self.J, self.mesh_normals.T)

    def __str__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.get_Jr()} (T)\n"
            + f"Center {self.get_center()}\n"
            + f"Orientation alpha,beta,gamma: {self.get_orientation()}\n"
        )
        return str

    def __repr__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.get_Jr()} (T)\n"
            + f"Center {self.get_center()}\n"
            + f"Orientation alpha,beta,gamma: {self.get_orientation()}\n"
        )
        return str

    def get_Jr(self):
        """Returns Magnetisation vector

        Returns:
            ndarray: [Jx, Jy, Jz]
        """
        return self.J

    def size(self):
        """Returns magnet dimesions

        Returns:
            size (ndarray): numpy array [width, depth, height]
        """
        pass

    def get_center(self):
        """Returns magnet center

        Returns:
            ndarray: magnet center
        """
        return self.center

    def get_field(self, x, y, z):
        """Calculates the magnetic field at point(s) x,y,z due to a 3D magnet
        The calculations are always performed in local coordinates with the centre of the magnet at origin and z magnetisation pointing along the local z' axis.

        The rotations and translations are performed first, and the internal field calculation functions are called.

        Args:
            x (float/array): x co-ordinates
            y (float/array): y co-ordinates
            z (float/array): z co-ordinates

        Returns:
            tuple: Bx(ndarray), By(ndarray), Bz(ndarray)  field vector
        """
        B = self._get_field_internal(x, y, z)

        return B.x, B.y, B.z

    def get_force_torque(self, depth=4, unit="mm"):
        """Calculates the force and torque on a prism magnet due to all other magnets.

        Args:
            depth (int, optional): Number of recursions of division by 4 per simplex
            unit (str, optional): Length scale. Defaults to 'mm'.

        Returns:
            tuple: force (ndarray (3,) ) and torque (ndarray (3,) )
        """
        from ..forces._mesh_force import calc_force_mesh

        force, torque = calc_force_mesh(self, depth, unit)
        return force, torque

    def _get_field_internal(self, x, y, z):
        """Internal magnetic field calculation methods.
        Iterates over each triangle that makes up the mesh magnet and calculates the magnetic field

        Args:
            x (float/array): x co-ordinates
            y (float/array): y co-ordinates
            z (float/array): z co-ordinates

        Returns:
            Field3: Magnetic field array
        """
        from ..utils._routines3D import _allocate_field_array3

        B = _allocate_field_array3(x, y, z)
        vec_shape = B.x.shape
        B.x = B.x.ravel()
        B.y = B.y.ravel()
        B.z = B.z.ravel()

        # debug for loop, used when needing to check certain triangles, or groups of triangles
        # for i in range(self.start, self.stop):
        for i in range(len(self.mesh_vectors)):
            if _np.fabs(self.Jnorm[i] / self.Jr) > 1e-4:
                Btx, Bty, Btz, _, _, _ = self.calcB_triangle(
                    self.mesh_vectors[i],
                    self.Jnorm[i],
                    x,
                    y,
                    z,
                    i,
                )

                B.x += Btx
                B.y += Bty
                B.z += Btz

        B.x = _np.reshape(B.x, vec_shape)
        B.y = _np.reshape(B.y, vec_shape)
        B.z = _np.reshape(B.z, vec_shape)

        B.n = _np.linalg.norm([B.x, B.y, B.z], axis=0)
        return B

    def _import_mesh(self):
        """Imports mesh from STL file

        Returns:
            tuple: mesh_vectors (ndarray of mesh triangles), mesh_normals (ndarray of normals to each triangle)
        """
        stl_mesh = mesh.Mesh.from_file(self._filename)

        if _np.any(
            _np.fabs([self.alpha_rad, self.beta_rad, self.gamma_rad]) > ALIGN_CUTOFF
        ):
            mesh_rotation = Quaternion.gen_rotation_quaternion(
                self.alpha_rad, self.beta_rad, self.gamma_rad
            )

            angle, axis = mesh_rotation.get_axisangle()
            stl_mesh.rotate(axis, angle)

        # to ensure that the initial center is set to the centroid
        _, centroid, _ = stl_mesh.get_mass_properties()
        stl_mesh.translate(-centroid)

        offset = self.get_center()
        stl_mesh.translate(offset / self.mesh_scale)

        # get values after translation
        volume, centroid, _ = stl_mesh.get_mass_properties()

        mesh_vectors = stl_mesh.vectors.astype(_np.float64)
        mesh_normals = stl_mesh.normals.astype(_np.float64)

        # scale values
        volume *= self.mesh_scale ** 3
        centroid *= self.mesh_scale
        mesh_vectors *= self.mesh_scale

        mesh_normals = mesh_normals / _np.linalg.norm(
            mesh_normals, axis=1, keepdims=True
        )

        return mesh_vectors, mesh_normals, volume, centroid

    def _generate_mask(self, x, y, z):
        """Generates mask of points inside a magnet
        NOTE: not implemented for Mesh magnets.
        Args:
            x (ndarray/float): x-coordinates
            y (ndarray/float): y-coordinates
            z (ndarray/float): z-coordinates
        """
        pass

    def calcB_triangle(self, triangle, Jr, x, y, z, i):
        """Calculates the magnetic field due to a triangle

        Args:
            triangle (ndarray): Vertices of a triangle
            Jr (float): Remnant magnetisation component normal to triangle
            x (ndarray): x coordinates
            y (ndarray): y coordinates
            z (ndarray): z coordinates

        Returns:
            tuple: Bx, By, Bz magnetic field components
        """

        (
            total_rotation,
            rotated_triangle,
            offset,
            RA_triangle1,
            RA_triangle2,
        ) = _rotate_triangle(triangle, Jr)

        # Prepare points and quaternion
        pos_vec = Quaternion._prepare_vector(x, y, z)

        # Rotate points
        x_rot, y_rot, z_rot = total_rotation * pos_vec

        norm1 = norm_plane(triangle)

        if _np.allclose(norm1, [0, -1, 0], atol=ALIGN_CUTOFF) and Jr < 0:
            RA_triangle1, RA_triangle2 = RA_triangle2, RA_triangle1

        Btx, Bty, Btz = self._calcB_2_triangles(
            RA_triangle1,
            RA_triangle2,
            Jr,
            x_rot - offset[0],
            y_rot - offset[1],
            z_rot - offset[2],
        )

        Bvec = Quaternion._prepare_vector(Btx, Bty, Btz)
        Bx, By, Bz = total_rotation.get_conjugate() * Bvec

        return Bx, By, Bz, rotated_triangle, offset, total_rotation

    def _calcB_2_triangles(self, triangle1, triangle2, Jr, x, y, z):
        """Calculates the magnetic field due to two split right angled triangles
        in their local frame.

        Args:
            triangle1 (ndarray): Vertices of triangle 1
            triangle2 (ndarray): Vertices of triangle 2
            Jr (float): normal remnant magnetisation
            x (ndarray): x coordinates
            y (ndarray): y coordinates
            z (ndarray): z coordinates

        Returns:
            tuple: Bx, By, Bz magnetic field components
        """

        # Calc RA1 Field
        Btx, Bty, Btz = self._charge_sheet(triangle1[0], triangle1[1], Jr, x, y, z)

        # Rotate into local of RA2
        rotate_about_z = Quaternion.q_angle_from_axis(PI, (0, 0, 1))
        pos_vec_RA2 = Quaternion._prepare_vector(x - triangle1[0], y, z)

        x_local, y_local, z_local = rotate_about_z * pos_vec_RA2

        # Calc RA2 Field
        Btx2, Bty2, Btz2 = self._charge_sheet(
            triangle2[0], triangle2[1], Jr, x_local + triangle2[0], y_local, z_local
        )

        # Inverse Rot of RA2 Field
        Bvec = Quaternion._prepare_vector(Btx2, Bty2, Btz2)
        Btx2, Bty2, Btz2 = rotate_about_z.get_conjugate() * Bvec

        Btx += Btx2
        Bty += Bty2
        Btz += Btz2

        return Btx, Bty, Btz

    @staticmethod
    def _charge_sheet(a, b, Jr, x, y, z):
        sigma = Jr
        with _np.errstate(all="ignore"):
            Bx = _charge_sheet_x(a, b, sigma, x, y, z)
            By = _charge_sheet_y(a, b, sigma, x, y, z)
            Bz = _charge_sheet_z(a, b, sigma, x, y, z)
        return Bx, By, Bz


@vectorize(
    [float64(float64, float64, float64, float64, float64, float64)], target="parallel"
)
def _charge_sheet_x(a, b, sigma, x, y, z):
    """Calculates the x-component of the magnetic field of a right angled charge sheet

    Args:
        a (float): triangle base
        b (float): triangle altitude
        sigma (float): normal magnetic charge density in tesla
        x (ndarray): x-coordinates
        y (ndarray): y-coordinates
        z (ndarray): z-coordinates

    Returns:
        ndarray: Bx magnetic field component
    """
    S_ab = 1.0 / sqrt(a ** 2 + b ** 2)

    r1 = sqrt(x ** 2 + y ** 2 + z ** 2)

    ax = a - x
    bz = b - z

    r2 = sqrt(ax ** 2 + y ** 2 + bz ** 2)
    r3 = sqrt(ax ** 2 + y ** 2 + z ** 2)
    t1_log = r1 - (a * x + b * z) * S_ab

    t2_log = r2 + (a * ax + b * bz) * S_ab

    # To avoid introduction of noise when taking log of values
    if fabs(t1_log) > FP_CUTOFF:
        t1 = log(t1_log)
    else:
        t1 = 0.0

    if fabs(t2_log) > FP_CUTOFF:
        t2 = log(t2_log)
    else:
        t2 = 0.0

    dt = t1 - t2
    Bx = b * S_ab * dt

    r3_minus_z = r3 - z
    r2_plus_bz = r2 + bz

    # if statements are used to avoid singularities such as divide by zero and miminise noise
    if fabs(r3_minus_z) > 0.0:
        r2_over_r3 = r2_plus_bz / r3_minus_z
        if fabs(r2_over_r3) > FP_CUTOFF:
            Bx += log(r2_plus_bz / r3_minus_z)

    Bx *= sigma / PI / 4

    return Bx


@vectorize(
    [float64(float64, float64, float64, float64, float64, float64)], target="parallel"
)
def _charge_sheet_z(a, b, sigma, x, y, z):
    """Calculates the z-component of the magnetic field of a right angled charge sheet

    Args:
    a (float): triangle base
    b (float): triangle altitude
    sigma (float): normal magnetic charge density in tesla
    x (ndarray): x-coordinates
    y (ndarray): y-coordinates
    z (ndarray): z-coordinates

    Returns:
    ndarray: Bz magnetic field component
    """

    S_ab = 1.0 / sqrt(a ** 2 + b ** 2)

    r1 = sqrt(x ** 2 + y ** 2 + z ** 2)

    ax = a - x
    bz = b - z

    r2 = sqrt(ax ** 2 + y ** 2 + bz ** 2)
    r3 = sqrt(ax ** 2 + y ** 2 + z ** 2)

    t1_log = r1 - (a * x + b * z) * S_ab
    t2_log = r2 + (a * ax + b * bz) * S_ab

    # To avoid introduction of noise when taking log of values
    if fabs(t1_log) > FP_CUTOFF:
        t1 = log(t1_log)
    else:
        t1 = 0.0

    if fabs(t2_log) > FP_CUTOFF:
        t2 = log(t2_log)
    else:
        t2 = 0.0

    Bz = a * S_ab * (t2 - t1)

    r3_plus_ax = r3 + ax
    r1_minus_x = r1 - x

    # if statements are used to avoid singularities such as divide by zero and miminise noise
    if fabs(r3_plus_ax) > 0.0:
        r1_over_r3 = r1_minus_x / r3_plus_ax
        if fabs(r1_over_r3) > FP_CUTOFF:
            Bz += log(r1_over_r3)
    Bz *= sigma / PI / 4

    return Bz


@vectorize(
    [float64(float64, float64, float64, float64, float64, float64)], target="parallel"
)
def _charge_sheet_y(a, b, sigma, x, y, z):
    """Calculates the y-component of the magnetic field of a right angled charge sheet

    Args:
    a (float): triangle base
    b (float): triangle altitude
    sigma (float): normal magnetic charge density in tesla
    x (ndarray): x-coordinates
    y (ndarray): y-coordinates
    z (ndarray): z-coordinates

    Returns:
    ndarray: By magnetic field component
    """

    # if statements are used to avoid singularities such as divide by zero
    if fabs(y) > FP_CUTOFF:
        a_ab = sqrt(1 + b ** 2 / a ** 2)
        r1 = sqrt(x ** 2 + y ** 2 + z ** 2)

        ax = a - x

        r3 = sqrt(ax ** 2 + y ** 2 + z ** 2)

        g = b / a

        si_alpha = 1.0 / (a_ab ** 2)

        beta = -(x + z * g) * si_alpha
        gamma_sq = (r1 ** 2 * si_alpha) - (beta ** 2)
        if gamma_sq > 0.0:
            gamma = sqrt(gamma_sq)
        else:
            gamma = 0.0

        A = -gamma * g
        B = gamma * a_ab
        C = z + beta * g
        ABC_diff = B ** 2 - A ** 2 - C ** 2

        if fabs(gamma) > 0.0:
            t3 = (a + beta) / gamma
            t4 = beta / gamma
        else:
            t3 = 0.0
            t4 = 0.0

        if ABC_diff > 0.0:
            t1 = sqrt(ABC_diff)
            t2 = 1.0 / t1
        else:
            t1 = 0.0
            t2 = 0.0

        atan_1 = t2 * (C + (A + B) * (sqrt(1 + t3 * t3) + t3))
        atan_2 = t2 * (C + (A + B) * (sqrt(1 + t4 * t4) + t4))
        atan_y = atan_1 - atan_2
        atan_x = 1 + atan_1 * atan_2

        a_ab_t1 = a_ab * t1

        # To avoid undefined atan2 behaviour
        if fabs(atan_y) < FP_CUTOFF and fabs(atan_x) < FP_CUTOFF:
            By = 0.0
        elif a_ab_t1 > FP_CUTOFF:
            By = (y / (a_ab_t1)) * atan2(atan_y, atan_x)
        else:
            By = 0.0

        t2 = 1 / y
        atan_1 = t2 * (r3 + z + (x - a))
        atan_2 = t2 * (r1 + z + x)
        atan_y = atan_1 - atan_2
        atan_x = 1 + atan_1 * atan_2

        if fabs(atan_y) < FP_CUTOFF and fabs(atan_x) < FP_CUTOFF:
            By = 0.0
        else:
            By = By + atan2(atan_y, atan_x)

        By = sigma * By / PI / 2

    else:
        By = 0.0

    return By



"""A setuptools based setup module for pymagnet
"""

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "pypi_README.md").read_text(encoding="utf-8")


setup(
    name="pymagnet",
    version="0.3.5",
    description="User Friendly Magnetic Field Calculations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pdunne/pymagnet",
    author="Peter Dunne",
    author_email="peter.dunne@applied-magnetism.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="magnetism, magnetic fields",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6, <4",
    install_requires=["numpy", "numpy-stl", "matplotlib", "numba", "plotly"],
    project_urls={
        "Documentation": "https://pdunne.github.io/pymagnet/",
        "Bug Reports": "https://github.com/pdunne/pymagnet/issues",
        "Source": "https://github.com/pdunne/pymagnet/",
    },
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""Quaternion module

Implements quaternion multiplication for convenient rotation of vectors in 3D.

Example:
    Rotation of a vector about the x-axis:

        import numpy as np
        import pymagnet as pm
        vector1 = np.array([1,0,0])
        rotate_about_z = pm.magnets.Quaternion.q_angle_from_axis(np.pi/2, (0, 0, 1))
        vector2 = rotate_about_z * vector1

"""
__all__ = ["Quaternion"]

import numpy as _np
from ..utils.global_const import FP_CUTOFF, MAG_TOL


class Quaternion:
    """Quaternion class.
    overloading of multiplication symbol allows easy quaternion multiplications

    """

    def __init__(self, w=1.0, x=0.0, y=0.0, z=0.0):
        """Initialse a pure quaternion (1; 0, 0, 0)

        Args:
            w (ndarray, optional): scalar quaternion. Defaults to 1.0.
            x (ndarray, optional): vector component. Defaults to 0.0.
            y (ndarray, optional): vector component. Defaults to 0.0.
            z (ndarray, optional): vector component. Defaults to 0.0.
        """
        self.w = _np.asarray(w)
        self.x = _np.asarray(x)
        self.y = _np.asarray(y)
        self.z = _np.asarray(z)

    def q_angle_from_axis(theta, vec):
        """Generates a rotation quaternion for an angle `theta` about an axis `vec`
        This is a normailsed, i.e. unit quaternion.

        Args:
            theta (float): angle of rotation
            vec (tuple/array): axis vector

        Example:
            90 degree rotation about the x axis:

                rotation_quaternion = Quaternion.q_angle_from_axis(np.pi/2, (1, 0, 0) )

        Returns:
            Quaternion: rotation quaternion
        """
        vec = Quaternion._normalise_axis(vec)
        w = _np.cos(theta / 2.0)
        vec *= _np.sin(theta / 2.0)
        x = vec[0]
        y = vec[1]
        z = vec[2]
        rotation_quaternion = Quaternion(w, x, y, z)
        return rotation_quaternion

    def get_conjugate(self):
        """Returns quaternion conjugate

        Returns:
            quaternion: quaternion conjugate
        """
        return Quaternion(self.w, -self.x, -self.y, -self.z)

    @staticmethod
    def gen_rotation_quaternion(alpha_rad=0.0, beta_rad=0.0, gamma_rad=0.0):
        """Generates quaternion for rotation around z, y, x axes

        Args:
            alpha_rad (float): angle to z-axis. Defaults to 0.0.
            beta_rad (float): angle to y-axis. Defaults to 0.0.
            gamma_rad (float): angle to x-axis. Defaults to 0.0.

        Returns:
            Quaternion: rotation quaternion
        """

        rotate_about_x = Quaternion()
        rotate_about_y = Quaternion()
        rotate_about_z = Quaternion()

        forward_rotation = Quaternion()

        if _np.fabs(alpha_rad) > MAG_TOL:
            rotate_about_z = Quaternion.q_angle_from_axis(alpha_rad, (0, 0, 1))

        if _np.fabs(beta_rad) > MAG_TOL:
            rotate_about_y = Quaternion.q_angle_from_axis(beta_rad, (0, 1, 0))

        if _np.fabs(gamma_rad) > MAG_TOL:
            rotate_about_x = Quaternion.q_angle_from_axis(gamma_rad, (1, 0, 0))

        forward_rotation = rotate_about_x * rotate_about_z * rotate_about_y

        return forward_rotation

    @staticmethod
    def _prepare_vector(x, y, z):
        """Creates 3xN array where each column corresponds to x, y, z vectors
        all of the same length. If some vectors are shorter, their values are repeated
        to ensure and array of (x,y,z) points for quaternion rotation.

        Args:
            x (ndarray): x coordinates
            y (ndarray): y coordinates
            z (ndarray): z coordinates

        Returns:
            ndarray: 3xN numpy ndarray
        """
        # Check input x,y,z are numpy arrays and if not, convert to numpy arrays
        x = _np.asarray(x)
        y = _np.asarray(y)
        z = _np.asarray(z)

        longest_array_length = _np.max([x.size, y.size, z.size])

        # Ensure the unravelled arrays are all of the same length
        x = Quaternion._check_extend_array(x.ravel(), longest_array_length)
        y = Quaternion._check_extend_array(y.ravel(), longest_array_length)
        z = Quaternion._check_extend_array(z.ravel(), longest_array_length)

        return _np.array([x, y, z])

    @staticmethod
    def _check_extend_array(array, max):
        """Extend a 1D vector to length 'max' if it is shorter than 'max'

        Args:
            array (ndarray): numpy array
            max (int): array length to compare to

        Returns:
            ndarray: numpy array of length 'max'
        """

        if max % array.size != 0:
            raise ValueError("Incorrect gridding of x,y,z")
        # If array is smaller than longest array length 'max', extend it by tiling
        if array.size < max:
            extended_array = _np.tile(array, (max // array.size))

            # # If there is a remainder, 'n' append the first n elements of the array
            # # to ensure the final length is 'max'.
            # if max % array.size != 0:
            #     extended_array = _np.append(
            #         extended_array, array[0 : (max % array.size)]
            #     )
            return extended_array
        else:
            return array

    @staticmethod
    def _normalise_axis(vec):
        """Normalise

        Args:
            v (array): [description]

        Returns:
            [type]: [description]
        """
        vec = _np.asarray(vec)
        if _np.fabs(_np.linalg.norm(vec, axis=0)) < FP_CUTOFF:
            raise ValueError("Vec norm should be non-zero")
        return vec / _np.linalg.norm(vec, axis=0)

    @staticmethod
    def vec_norm(x, y, z):
        """Normalises each x,y,z vector

        Args:
            x (ndarray): x array
            y (ndarray): y array
            z (ndarray): z array

        Returns:
            array: 3xN array of normalised vectors
        """
        vec = Quaternion._prepare_vector(x, y, z)
        return _np.linalg.norm(vec, axis=0)

    def __repr__(self):
        str = f"({self.w}; {self.x}, {self.y}, {self.z})"
        return str

    def __mul__(self, b):
        if isinstance(b, Quaternion):
            return self._multiply_with_quaternion(b)
        elif isinstance(b, (list, tuple, _np.ndarray)):
            if len(b) != 3:
                raise Exception(f"Input vector has invalid length {len(b)}")
            return self._multiply_with_vector(b)
        else:
            raise Exception(f"Multiplication with unknown type {type(b)}")

    def as_tuple(self):
        """Returns quaternion as tuple of arrays

        Returns:
            tuple: w (array), x (array), y (array), z (array)
        """
        return self.w, self.x, self.y, self.z

    def _multiply_with_quaternion(self, q2):
        """Computes the Hamilton product of two quaternions

        Args:
            q2 (quaternion): quaternion

        Returns:
            quaternion: Hamilton product
        """
        w1, x1, y1, z1 = self.as_tuple()
        w2, x2, y2, z2 = q2.as_tuple()
        w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
        x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
        y = w1 * y2 - x1 * z2 + y1 * w2 + z1 * x2
        z = w1 * z2 + x1 * y2 - y1 * x2 + z1 * w2
        result = Quaternion(w, x, y, z)
        return result

    def _multiply_with_vector(self, v):
        """Converts vector ``v`` to quaternion ``q2`` and then performs rotation by
        multiplying `q * q2 * q'`


        Args:
            v (vector/array): vector to be rotated

        Returns:
            tuple: multiplied vector x', y', z'
        """
        q2 = Quaternion(_np.zeros_like(v[0]), v[0], v[1], v[2])
        result = self * q2 * self.get_conjugate()
        return result.x, result.y, result.z

    def get_axisangle(self):
        theta = _np.arccos(self.w) * 2.0
        vec = _np.hstack([self.x, self.y, self.z])
        return theta, self._normalise_axis(vec)

    @staticmethod
    def euler_to_quaternion(alpha, beta, gamma):
        """Converts Euler angles to quaternion

        Args:
            alpha (float): angle to z-axis
            beta (float): angle to y-axis
            gamma (float): angle to x-axis

        Returns:
            Quaternion: Euler angles as a Quaternion
        """

        qw = _np.cos(alpha / 2) * _np.cos(beta / 2) * _np.cos(gamma / 2) + _np.sin(
            alpha / 2
        ) * _np.sin(beta / 2) * _np.sin(gamma / 2)
        qx = _np.sin(alpha / 2) * _np.cos(beta / 2) * _np.cos(gamma / 2) - _np.cos(
            alpha / 2
        ) * _np.sin(beta / 2) * _np.sin(gamma / 2)
        qy = _np.cos(alpha / 2) * _np.sin(beta / 2) * _np.cos(gamma / 2) + _np.sin(
            alpha / 2
        ) * _np.cos(beta / 2) * _np.sin(gamma / 2)
        qz = _np.cos(alpha / 2) * _np.cos(beta / 2) * _np.sin(gamma / 2) - _np.sin(
            alpha / 2
        ) * _np.sin(beta / 2) * _np.cos(gamma / 2)

        return Quaternion(qw, qx, qy, qz)


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""pymagnets.magnets

This module imports the classes and functions in the private modules to create a public API. 

"""
from ._magnet_base import *
from ._magnet1D import *
from ._magnet2D import *
from ._magnet3D import *
from ._polygon2D import *
from ._polygon3D import *


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""Plotting routines for calculating along symmetry lines of cubes, cuboids, and cylinders

"""
import matplotlib.pyplot as _plt
import numpy as _np
from ..utils._vector_structs import Point_Array1

from ..magnets import (
    Cylinder,
    Prism,
    magnetic_field_prism_1D,
    magnetic_field_cylinder_1D,
)

__all__ = ["plot_1D_field"]


def plot_1D_field(magnet, unit="mm", **kwargs):
    """Calculates and plots the magnetic field along the central symmetry axis
    of a cylinder or cuboid magnet, assuming the magnetic field is collinear

    Args:
        magnet (Magnet3D): Must be a Magnet3D type of magnet, either Prism, Cube,or Cylinder.

    Kwargs:
        num_points (int): Number of points to calculate. Defaults to 101.

    Returns:
        tuple: Point_Array1, Field1: point array struct containing z and the unit (e/g. 'mm'), vector array containing Bz and the field unit (e.g. 'T').
    """

    num_points = kwargs.pop("num_points", 101)
    return_data = kwargs.pop("return_data", False)
    points = Point_Array1(_np.zeros(num_points), unit=unit)

    if issubclass(magnet.__class__, Cylinder):
        mag_boundary = magnet.length / 2
        points.z = _np.linspace(
            -2 * magnet.length + magnet.center[2],
            2 * magnet.length + magnet.center[2],
            num_points,
        )
        field = magnetic_field_cylinder_1D(magnet, points.z)

        # if true, apply NaNs to inside the magnet
        if magnet._mask_magnet:
            mask = _generate_mask_1D(mag_boundary, magnet.center[2], points.z)
            field.z[mask] = _np.NaN

    elif issubclass(magnet.__class__, Prism):
        mag_boundary = magnet.height / 2
        points.z = _np.linspace(
            -2 * magnet.height + magnet.center[2],
            2 * magnet.height + magnet.center[2],
            num_points,
        )
        field = magnetic_field_prism_1D(magnet, points.z)

        # if true, apply NaNs to inside the magnet
        if magnet._mask_magnet:
            mask = _generate_mask_1D(mag_boundary, magnet.center[2], points.z)
            field.z[mask] = _np.NaN

    else:
        print("Error")
        return None

    fig, ax = _plt.subplots(figsize=(8, 8))
    unit_length = "(" + points.unit + ")"
    field_unit = "(" + field.unit + ")"
    _plt.xlabel(r"$z$ " + unit_length)
    _plt.ylabel(r"$B_z$ " + field_unit)
    _plt.plot(points.z, field.z)
    _plt.axvline(x=-mag_boundary + magnet.center[2], c="blue", ls="--")
    _plt.axvline(x=mag_boundary + magnet.center[2], c="red", ls="--")
    _plt.axvline(x=0.0, c="k", ls="-")
    _plt.show()

    if return_data:
        return points, field


def _generate_mask_1D(mag_boundary, zc, z):
    """Generates mask for points inside magnet

    Args:
        mag_boundary (float): half the height/length of a magnet
        zc (float): center of the magnet
        z (ndarray): z-coordinates

    Returns:
        ndarray: boolean mask array
    """
    zn = -mag_boundary + zc
    zp = mag_boundary + zc
    mask = _np.logical_and(z > zn, z < zp)

    return mask


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""pymagnets.plots

This module imports the classes and functions in the private modules to create a public API.

- Lines and contour plots are drawn using matplotlib
- 3D surface and volume plots are rendered using plotly.

"""
from ._plot1D import *
from ._plot2D import *
from ._plotly3D import *

# from ._plotly2D import *



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v.2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""Magnet Base class

This private module implements the registry and base magnet classes
"""
__all__ = ["Magnet"]

from weakref import WeakSet
from ..utils.global_const import MAG_TOL
import numpy as _np


class Registry:
    """Registry class for tracking instances

    Instances are tracked in `class.instances` using weak references.
    This also includes any instances that are deleted manually or go out of
    scope.

    Class methods:

        `print_instances()`

        `get_instances()`

        `get_num_instances()`

        `reset()
    """

    instances = WeakSet()
    _class_instances = []

    def __new__(cls, *args, **kwargs):
        o = object.__new__(cls)
        cls._register_instance(o)
        return o

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.__class__._class_instances.append(self)

    def __del__(self) -> None:
        if self in self.__class__._class_instances:
            self.__class__._class_instances.remove(self)

    @classmethod
    def print_instances(cls):
        """Prints class instantiations"""
        if len(cls.instances) < 1:
            print("No Instances")
        else:
            for instance in cls.instances:
                print(instance)

    @classmethod
    def get_instances(cls):
        """Gets lists of instances

        Returns:
            set: all class instances
        """
        return cls.instances

    @classmethod
    def get_num_instances(cls, Print_Val=False):

        """Return number of instances of class

        Args:
            Print_Val (bool, optional): [Print to screen]. Defaults to False.

        Returns:
            num_instances [int]:
        """
        if Print_Val:
            print(len(cls.instances))
        return len(cls.instances)

    @classmethod
    def _register_instance(cls, instance):
        """Adds class instance to registry

        Args:
            instance (instance): class instance
        """
        cls.instances.add(instance)
        for b in cls.__bases__:
            if issubclass(b, Registry):
                b._register_instance(instance)

    @classmethod
    def reset(cls):
        """Removes all instances from registry."""
        for magnet in cls._class_instances:
            del magnet
        cls.instances = WeakSet()
        cls._class_instances = []

    def __init_subclass__(cls):
        cls.instances = WeakSet()
        cls._class_instances = []


class Magnet(Registry):
    """Magnet base class


    Returns:
        Magnet: magnet base class
    """

    tol = MAG_TOL  # tolerance for rotations, sufficient for 0.01 degree accuracy
    mag_type = "Magnet"

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.center = _np.array([0.0, 0.0])


def reset_magnets():
    """Returns a list of all instantiated magnets."""
    from ._magnet2D import Magnet2D, Rectangle, Square, Circle
    from ._polygon2D import PolyMagnet
    from ._magnet3D import Magnet3D, Prism, Cube, Cylinder, Sphere
    from ._polygon3D import Mesh

    magnet_classes = [
        # Registry,
        Magnet,
        Magnet2D,
        Rectangle,
        Square,
        Circle,
        PolyMagnet,
        Magnet3D,
        Prism,
        Cube,
        Cylinder,
        Sphere,
        Mesh,
    ]
    for cls in magnet_classes:
        cls.reset()


def list_magnets():
    """Returns a list of all instantiated magnets.

    Assumes that the child class registries have not been modified outside of
    using `pymagnet.reset_magnets()`.
    """
    return Magnet.print_instances()
